# This is a quick hack to convert an openldap schema file to a form which
# can be loaded into Samba4/AD.
#
# Inspired by:
# http://david-latham.blogspot.co.uk/2012/12/extending-ad-schema-on-samba4-part-2.html
# https://github.com/linuxplayground/yubikey-ldap/tree/master/samba4-schema
#
# (c) 2017 Brian Candler <b.candler@pobox.com>
# -------------------------------------------------------------------------
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
# -----------------------------------------------------------------------

from __future__ import print_function
import sys
import re
from collections import OrderedDict

BASEDN = 'dc=samba4,dc=internal'

# RFC 2252 to https://technet.microsoft.com/en-us/library/cc961740.aspx
SYNTAX_MAP = {
    '1.3.6.1.4.1.1466.115.121.1.7':  ('2.5.5.8',   1), # boolean
    '1.3.6.1.4.1.1466.115.121.1.12': ('2.5.5.1', 127), # DN
    '1.3.6.1.4.1.1466.115.121.1.15': ('2.5.5.3',  27), # DirectoryString
    '1.3.6.1.4.1.1466.115.121.1.26': ('2.5.5.5',  22), # IA5String
    '1.3.6.1.4.1.1466.115.121.1.27': ('2.5.5.9',  10), # Integer
}
obj = None
for line in sys.stdin:
    if re.match(r'^\s*(#|$)', line): continue
    m = re.match(r'^attributetype\s+\(\s+(\S+)', line)
    if m:
        obj = OrderedDict([
            ('objectClass', ['top', 'attributeSchema']),
            ('attributeID', m.group(1)),
            ('isSingleValued', 'FALSE'),
        ])
        continue
    m = re.match(r'^objectclass\s+\(\s+(\S+)', line)
    if m:
        obj = OrderedDict([
            ('objectClass', ['top', 'classSchema']),
            ('governsID', m.group(1)),
        ])
        continue
    m = re.match(r'^\s*NAME\s+[\'"](.+)[\'"]', line)
    if m:
        obj.update([
            ('cn', m.group(1)),
            ('name', m.group(1)),
            ('lDAPDisplayName', m.group(1)),
        ])
        continue
    m = re.match(r'^\s*DESC\s+[\'"](.+)[\'"]', line)
    if m:
        obj.update([
            ('description', m.group(1)),
        ])
        continue
    m = re.match(r'^\s*(EQUALITY|SUBSTR)\s+(\S+)', line)
    if m:
        # Not supported by AD?
        # https://technet.microsoft.com/en-us/library/cc961575.aspx
        continue
    m = re.match(r'^\s*SYNTAX\s+(\S+)', line)
    if m:
        obj.update([
            ('attributeSyntax', SYNTAX_MAP[m.group(1)][0]),
            ('oMSyntax', SYNTAX_MAP[m.group(1)][1]),
        ])
        continue
    if re.match(r'^\s*SINGLE-VALUE', line):
        obj.update([
            ('isSingleValued', 'TRUE'),
        ])
        continue
    if re.match(r'^\s*AUXILIARY', line):
        # https://msdn.microsoft.com/en-us/library/ms679014(v=vs.85).aspx
        # https://technet.microsoft.com/en-us/library/2008.05.schema.aspx
        obj.update([
            ('objectClassCategory', '3'),
        ])
        continue
    if re.match(r'^\s*STRUCTURAL', line):
        obj.update([
            ('objectClassCategory', '1'),
        ])
        continue
    m = re.match(r'^\s*SUP\s+(\S+)', line)
    if m:
        obj.update([
            ('subClassOf', m.group(1)),
        ])
        continue
    m = re.match(r'^\s*(MAY|MUST)\s+\((.*)\)\s*$', line)
    if m:
        attrs = m.group(2).split('$')
        obj.update([
            ('%sContain' % m.group(1).lower(), [v.strip() for v in attrs]),
        ])
        continue
    m = re.match(r'^\s*(MAY|MUST)\s+(\w+)\s*$', line)
    if m:
        obj.update([
            ('%sContain' % m.group(1).lower(), m.group(2)),
        ])
        continue
    if re.match(r'^\s*\)', line) and obj:
        print("dn: CN=%s,CN=Schema,CN=Configuration,%s" % (obj['cn'], BASEDN))
        print("changetype: add")
        for k in obj:
            if type(obj[k]) == list:
                for v in obj[k]:
                    print("%s: %s" % (k, v))
            else:
                print("%s: %s" % (k, obj[k]))
        print()
        obj = None
        continue
    print("??? %s" % line, file=sys.stderr)



#! /usr/bin/env python3
#
# Example Python module for prepaid usage using MySQL

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
#
# Copyright 2002 Miguel A.L. Paraz <mparaz@mparaz.com>
# Copyright 2002 Imperium Technology, Inc.
#
# $Id: 28e548b47cbf63ca55f1445cd1f5f0adf0bd436a $

import radiusd
import MySQLdb

# Configuration
configDb = 'python'                     # Database name
configHost = 'localhost'                # Database host
configUser = 'python'                   # Database user and password
configPasswd = 'python'

# xxx Database

# Globals
dbHandle = None

def log(level, s):
    """Log function."""
    radiusd.radlog(level, 'prepaid.py: ' + s)

def instantiate(p):
    """Module Instantiation.  0 for success, -1 for failure.
    p is a dummy variable here."""
    global dbHandle

    try:
    	dbHandle = MySQLdb.connect(db=configDb, host=configHost,
    				   user=configUser, passwd=configPasswd)

    except MySQLdb.OperationalError as e:
    	# Report the error and return -1 for failure.
    	# xxx A more advanced module would retry the database.
    	log(radiusd.L_ERR, str(e))
    	return -1

    log(radiusd.L_INFO, 'db connection: ' + str(dbHandle))

    return 0


def authorize(authData):
    """Authorization and authentication are done in one step."""

    # Extract the data we need.
    userName = None
    userPasswd = None

    for t in authData:
    	if t[0] == 'User-Name':
    		userName = t[1]
    	elif t[0] == 'Password':
    		userPasswd = t[1]

    # Build and log the SQL statement
    # radiusd puts double quotes (") around the string representation of
    # the RADIUS packet.
    sql = 'select passwd, maxseconds from users where username = ' +  userName

    log(radiusd.L_DBG, sql)

    # Get a cursor
    # xxx Or should this be one cursor all throughout?
    try:
    	dbCursor = dbHandle.cursor()
    except MySQLdb.OperationalError as e:
    	log(radiusd.L_ERR, str(e))
    	return radiusd.RLM_MODULE_FAIL

    # Execute the SQL statement
    try:
    	dbCursor.execute(sql)
    except MySQLdb.OperationalError as e:
    	log(radiusd.L_ERR, str(e))
    	dbCursor.close()
    	return radiusd.RLM_MODULE_FAIL

    # Get the result. (passwd, maxseconds)
    result = dbCursor.fetchone()
    if not result:
    	# User not found
    	log(radiusd.L_INFO, 'user not found: ' + userName)
    	dbCursor.close()
    	return radiusd.RLM_MODULE_NOTFOUND



    # Compare passwords
    # Ignore the quotes around userPasswd.
    if result[0] != userPasswd[1:-1]:
    	log(radiusd.L_DBG, 'user password mismatch: ' + userName)
    	return radiusd.RLM_MODULE_REJECT

    maxSeconds = result[1]

    # Compute their session limit

    # Build and log the SQL statement
    sql = 'select sum(seconds) from sessions where username = ' + userName

    log(radiusd.L_DBG, sql)

    # Execute the SQL statement
    try:
    	dbCursor.execute(sql)
    except MySQLdb.OperationalError as e:
    	log(radiusd.L_ERR, str(e))
    	dbCursor.close()
    	return radiusd.RLM_MODULE_FAIL

    # Get the result. (sum,)
    result = dbCursor.fetchone()
    if (not result) or (not result[0]):
    	# No usage yet
    	secondsUsed = 0
    else:
    	secondsUsed = result[0]

    # Done with cursor
    dbCursor.close()

    # Note that MySQL returns the result of SUM() as a float.
    sessionTimeout = maxSeconds - int(secondsUsed)

    if sessionTimeout <= 0:
    	# No more time, reject outright
    	log(radiusd.L_INFO, 'user out of time: ' + userName)
    	return radiusd.RLM_MODULE_REJECT

    # Log the success
    log(radiusd.L_DBG, 'user accepted: %s, %d seconds' %
    		(userName, sessionTimeout))

    # We are adding to the RADIUS packet
    # Note that the session timeout integer must be converted to string.
    # We need to set an Auth-Type.

    return (radiusd.RLM_MODULE_UPDATED,
    	(('Session-Timeout', str(sessionTimeout)),),
    	(('Auth-Type', 'python'),))
    # If you want to use different operators
    # you can do
    # return (radiusd.RLM_MODULE_UPDATED,
    #         (
    #            ('Session-Timeout', ':=', str(sessionTimeout)),
    #            ('Some-other-option', '-=', Value'),
    #         ),
    #         (
    #            ('Auth-Type', ':=', 'python'),
    #         ),
    #        )

def authenticate(p):
    return radiusd.RLM_MODULE_OK


def preacct(p):
    return radiusd.RLM_MODULE_OK


def accounting(acctData):
    """Accounting."""
    # Extract the data we need.

    userName = None
    acctSessionTime = None
    acctStatusType = None

    # xxx A dict would make this nice.
    for t in acctData:
    	if t[0] == 'User-Name':
    		userName = t[1]
    	elif t[0] == 'Acct-Session-Time':
    		acctSessionTime = t[1]
    	elif t[0] == 'Acct-Status-Type':
    		acctStatusType = t[1]


    # We will not deal with Start for now.
    # We may later, for simultaneous checks and the like.
    if acctStatusType == 'Start':
    	return radiusd.RLM_MODULE_OK

    # Build and log the SQL statement
    # radiusd puts double quotes (") around the string representation of
    # the RADIUS packet.
    #
    # xxx This is simplistic as it does not record the time, etc.
    #
    sql = 'insert into sessions (username, seconds) values (%s, %d)' % \
    (userName, int(acctSessionTime))

    log(radiusd.L_DBG, sql)

    # Get a cursor
    # xxx Or should this be one cursor all throughout?
    try:
    	dbCursor = dbHandle.cursor()
    except MySQLdb.OperationalError as e:
    	log(radiusd.L_ERR, str(e))
    	return radiusd.RLM_MODULE_FAIL

    # Execute the SQL statement
    try:
    	dbCursor.execute(sql)
    except MySQLdb.OperationalError as e:
    	log(radiusd.L_ERR, str(e))
    	dbCursor.close()
    	return radiusd.RLM_MODULE_FAIL


    return radiusd.RLM_MODULE_OK


def detach():
    """Detach and clean up."""
    # Shut down the database connection.
    global dbHandle
    log(radiusd.L_DBG, 'closing database handle: ' + str(dbHandle))
    dbHandle.close()

    return radiusd.RLM_MODULE_OK



# Test the modules
if __name__ == '__main__':
    instantiate(None)
    print(authorize((('User-Name', '"map"'), ('User-Password', '"abc"'))))



#! /usr/bin/env python2
#
# Python module test
# Miguel A.L. Paraz <mparaz@mparaz.com>
#
# $Id: 97b5b64f088b77ee4be289a369ce6b432254c1e4 $

import radiusd

def instantiate(p):
  print "*** instantiate ***"
  print p

def authorize(p):
  print "*** authorize ***"
  print
  radiusd.radlog(radiusd.L_INFO, '*** radlog call in authorize ***')
  print
  print p 
  return radiusd.RLM_MODULE_OK

def preacct(p):
  print "*** preacct ***"
  print p 
  return radiusd.RLM_MODULE_OK

def accounting(p):
  print "*** accounting ***"
  radiusd.radlog(radiusd.L_INFO, '*** radlog call in accounting (0) ***')
  print
  print p 
  return radiusd.RLM_MODULE_OK

def pre_proxy(p):
  print "*** pre_proxy ***"
  print p 
  return radiusd.RLM_MODULE_OK

def post_proxy(p):
  print "*** post_proxy ***"
  print p 
  return radiusd.RLM_MODULE_OK

def post_auth(p):
  print "*** post_auth ***"
  print p 
  return radiusd.RLM_MODULE_OK

def recv_coa(p):
  print "*** recv_coa ***"
  print p 
  return radiusd.RLM_MODULE_OK

def send_coa(p):
  print "*** send_coa ***"
  print p 
  return radiusd.RLM_MODULE_OK


def detach():
  print "*** goodbye from radiusd_test.py ***"
  return radiusd.RLM_MODULE_OK




#! /usr/bin/env python2
#
# Example Python module for prepaid usage using MySQL

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
#
# Copyright 2002 Miguel A.L. Paraz <mparaz@mparaz.com>
# Copyright 2002 Imperium Technology, Inc.
#
# $Id: 3b1dc2e2e8e1456ae917c414b515b4f1f0df7a65 $

import radiusd
import MySQLdb

# Configuration
configDb = 'python'                     # Database name
configHost = 'localhost'                # Database host
configUser = 'python'                   # Database user and password
configPasswd = 'python'

# xxx Database

# Globals
dbHandle = None

def log(level, s):
  """Log function."""
  radiusd.radlog(level, 'prepaid.py: ' + s)

def instantiate(p):
  """Module Instantiation.  0 for success, -1 for failure.
  p is a dummy variable here."""
  global dbHandle

  p = p

  try:
    dbHandle = MySQLdb.connect(db=configDb, host=configHost,
			       user=configUser, passwd=configPasswd)

  except MySQLdb.OperationalError, e:
    # Report the error and return -1 for failure.
    # xxx A more advanced module would retry the database.
    log(radiusd.L_ERR, str(e))
    return -1

  log(radiusd.L_INFO, 'db connection: ' + str(dbHandle))

  return 0


def authorize(authData):
  """Authorization and authentication are done in one step."""

  # Extract the data we need.
  userName = None
  userPasswd = None

  for t in authData:
    if t[0] == 'User-Name':
      userName = t[1]
    elif t[0] == 'Password':
      userPasswd = t[1]

  # Build and log the SQL statement
  # radiusd puts double quotes (") around the string representation of
  # the RADIUS packet.
  sql = 'select passwd, maxseconds from users where username = ' +  userName

  log(radiusd.L_DBG, sql)

  # Get a cursor
  # xxx Or should this be one cursor all throughout?
  try:
    dbCursor = dbHandle.cursor()
  except MySQLdb.OperationalError, e:
    log(radiusd.L_ERR, str(e))
    return radiusd.RLM_MODULE_FAIL

  # Execute the SQL statement
  try:
    dbCursor.execute(sql)
  except MySQLdb.OperationalError, e:
    log(radiusd.L_ERR, str(e))
    dbCursor.close()
    return radiusd.RLM_MODULE_FAIL

  # Get the result. (passwd, maxseconds)
  result = dbCursor.fetchone()
  if not result:
    # User not found
    log(radiusd.L_INFO, 'user not found: ' + userName)
    dbCursor.close()
    return radiusd.RLM_MODULE_NOTFOUND



  # Compare passwords
  # Ignore the quotes around userPasswd.
  if result[0] != userPasswd[1:-1]:
    log(radiusd.L_DBG, 'user password mismatch: ' + userName)
    return radiusd.RLM_MODULE_REJECT

  maxSeconds = result[1]

  # Compute their session limit

  # Build and log the SQL statement
  sql = 'select sum(seconds) from sessions where username = ' + userName

  log(radiusd.L_DBG, sql)

  # Execute the SQL statement
  try:
    dbCursor.execute(sql)
  except MySQLdb.OperationalError, e:
    log(radiusd.L_ERR, str(e))
    dbCursor.close()
    return radiusd.RLM_MODULE_FAIL

  # Get the result. (sum,)
  result = dbCursor.fetchone()
  if (not result) or (not result[0]):
    # No usage yet
    secondsUsed = 0
  else:
    secondsUsed = result[0]

  # Done with cursor
  dbCursor.close()

  # Note that MySQL returns the result of SUM() as a float.
  sessionTimeout = maxSeconds - int(secondsUsed)

  if sessionTimeout <= 0:
    # No more time, reject outright
    log(radiusd.L_INFO, 'user out of time: ' + userName)
    return radiusd.RLM_MODULE_REJECT

  # Log the success
  log(radiusd.L_DBG, 'user accepted: %s, %d seconds' %
      (userName, sessionTimeout))

  # We are adding to the RADIUS packet
  # Note that the session timeout integer must be converted to string.
  # We need to set an Auth-Type.

  return (radiusd.RLM_MODULE_UPDATED,
	  (('Session-Timeout', str(sessionTimeout)),),
	  (('Auth-Type', 'python'),))
  # If you want to use different operators
  # you can do
  # return (radiusd.RLM_MODULE_UPDATED,
  #         (
  #            ('Session-Timeout', ':=', str(sessionTimeout)),
  #            ('Some-other-option', '-=', Value'),
  #         ),
  #         (
  #            ('Auth-Type', ':=', 'python'),
  #         ),
  #        )

def authenticate(p):
  p = p
  return radiusd.RLM_MODULE_OK


def preacct(p):
  p = p
  return radiusd.RLM_MODULE_OK


def accounting(acctData):
  """Accounting."""
  # Extract the data we need.

  userName = None
  acctSessionTime = None
  acctStatusType = None

  # xxx A dict would make this nice.
  for t in acctData:
    if t[0] == 'User-Name':
      userName = t[1]
    elif t[0] == 'Acct-Session-Time':
      acctSessionTime = t[1]
    elif t[0] == 'Acct-Status-Type':
      acctStatusType = t[1]


  # We will not deal with Start for now.
  # We may later, for simultaneous checks and the like.
  if acctStatusType == 'Start':
    return radiusd.RLM_MODULE_OK

  # Build and log the SQL statement
  # radiusd puts double quotes (") around the string representation of
  # the RADIUS packet.
  #
  # xxx This is simplistic as it does not record the time, etc.
  #
  sql = 'insert into sessions (username, seconds) values (%s, %d)' % \
	(userName, int(acctSessionTime))

  log(radiusd.L_DBG, sql)

  # Get a cursor
  # xxx Or should this be one cursor all throughout?
  try:
    dbCursor = dbHandle.cursor()
  except MySQLdb.OperationalError, e:
    log(radiusd.L_ERR, str(e))
    return radiusd.RLM_MODULE_FAIL

  # Execute the SQL statement
  try:
    dbCursor.execute(sql)
  except MySQLdb.OperationalError, e:
    log(radiusd.L_ERR, str(e))
    dbCursor.close()
    return radiusd.RLM_MODULE_FAIL


  return radiusd.RLM_MODULE_OK


def detach():
  """Detach and clean up."""
  # Shut down the database connection.
  global dbHandle
  log(radiusd.L_DBG, 'closing database handle: ' + str(dbHandle))
  dbHandle.close()

  return radiusd.RLM_MODULE_OK



# Test the modules
if __name__ == '__main__':
  instantiate(None)
  print authorize((('User-Name', '"map"'), ('User-Password', '"abc"')))
